/* mohawk - small HTTP server
**
** Copyright  1999,2000 by Jef Poskanzer <jef@mail.acme.com>.
** Copyright  2010 by Baptiste Daroussin <bapt@FreeBSD.org>,
** Copyright  2010 by Freddy Dissaux <freddy.dsx@free.fr>,
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
** OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
*/


#include "mohawk.h"

#include <unistd.h>
#include <stdlib.h>
#define _WITH_GETLINE
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <sys/stat.h>
#include <pwd.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <ctype.h>
#include <netinet/in.h>
#include <resolv.h>
#include <sys/wait.h>
#include <sys/event.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <dirent.h>
#include <err.h>
#include <fnmatch.h>
#include <sys/uio.h>
#include <stdint.h>

extern char *crypt(const char *key, const char *setting);

#if defined(AF_INET6) && defined(IN6_IS_ADDR_V4MAPPED)
#define USE_IPV6
#endif

#ifndef STDIN_FILENO
#define STDIN_FILENO 0
#endif
#ifndef STDOUT_FILENO
#define STDOUT_FILENO 1
#endif
#ifndef STDERR_FILENO
#define STDERR_FILENO 2
#endif

#ifndef SHUT_WR
#define SHUT_WR 1
#endif

#ifndef SIZE_T_MAX
#define SIZE_T_MAX 2147483647L
#endif

#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif
#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif


#ifndef ERR_DIR
#define ERR_DIR "errors"
#endif /* ERR_DIR */
#ifndef DEFAULT_HTTP_PORT
#define DEFAULT_HTTP_PORT "80"
#endif /* DEFAULT_HTTP_PORT */
#ifndef DEFAULT_USER
#define DEFAULT_USER "nobody"
#endif /* DEFAULT_USER */
#ifndef CGI_NICE
#define CGI_NICE 10
#endif /* CGI_NICE */
#ifndef CGI_PATH
#define CGI_PATH "/usr/local/bin:/usr/ucb:/bin:/usr/bin"
#endif /* CGI_PATH */
#ifndef CGI_LD_LIBRARY_PATH
#define CGI_LD_LIBRARY_PATH "/usr/local/lib:/usr/lib"
#endif /* CGI_LD_LIBRARY_PATH */
#ifndef AUTH_FILE
#define AUTH_FILE ".htpasswd"
#endif /* AUTH_FILE */
#ifndef READ_TIMEOUT
#define READ_TIMEOUT 60
#endif /* READ_TIMEOUT */
#ifndef WRITE_TIMEOUT
#define WRITE_TIMEOUT 300
#endif /* WRITE_TIMEOUT */
#ifndef DEFAULT_CHARSET
#define DEFAULT_CHARSET "iso-8859-1"
#endif /* DEFAULT_CHARSET */
#ifndef DEFAULT_PIDFILE
#define DEFAULT_PIDFILE "/var/run/mohawk.pid"
#endif /* DEFAULT_PIDFILE */


#define METHOD_UNKNOWN 0
#define METHOD_GET 1
#define METHOD_HEAD 2
#define METHOD_POST 3

#define DATE_1123 "%a, %d %b %Y %H:%M:%S"
#define DATE_1036 "%A, %d-%b-%y %H:%M:%S"
#define DATE_ANSI "%a %b %d %H:%M:%S %Y"

struct httpd conf;
struct vhost *vh;

/* Forwards. */
static void usage(void);
static void handle_request(struct client *);
static void do_file(struct client *, char *, struct stat *);
static void do_dir(struct client *, char *, struct stat *);
static void do_status(struct client *);
static void file_details(struct sbuf *, const char *dir, const char *name, const char *script_name);
static void strencode(char *to, size_t tosize, const char *from);
static void do_cgi(struct client *, char *);
static char **make_envp(struct client *);
static char *build_env(struct sbuf *, char *fmt, ...);
static int auth_check(struct client *, char *dirname);
static void send_authenticate(struct client *cl, char *realm);
static void send_error(struct client *, int s, char *title, char *extra_header, char *text);
static void add_headers(struct client *, int s, char *title, char *extra_header, char *me, char *mt, off_t b, time_t mod);
static ssize_t mohawk_read(int fd, char *buf, size_t size);
static ssize_t mohawk_write(int fd, char *buf, size_t size);
static char *get_method_str(int m);
static void handle_sigterm(int sig);
static void handle_sighup(int sig);
static void handle_sigchld(int sig);
static void parse_hdrs_line(struct client *, char *);
static void parse_cgi_headers(struct client *cl);
/*static void handle_read_timeout(int sig, struct client *);*/ /* FIXME */
/*static void handle_write_timeout(int sig);*/
static struct listener *get_listener_event(int fd);
static int read_http_header(struct client *, int, long);
static int read_cgi_header(struct client *, int, long);
static int read_cgi_data(struct client *, int, long);
static int read_client(struct client *cl, int fd, size_t len);
static int pass_cgi_stdin(struct client *cl, int fd, int len);
static int write_client_data(struct client *, int);
static struct client *client_new(int, const char *);
static void client_free(struct client *);
static int get_index(char *, char **);
time_t parse_date(const char *str);
static void mohawk_serve(void);

int
main(int argc, char** argv)
{
	/* new config */
	struct listener *listener;
	struct passwd* pwd = NULL;
	uid_t uid = 32767;
	gid_t gid = 32767;
	int ch;
	int fd;
	pid_t otherpid;
	int flag_config = 0;
	int nb_listen = 0;

	/* configuration initialization */
	SLIST_INIT(&conf.list);
	SLIST_INIT(&conf.vhosts);

	conf.servername = NULL;
	conf.chroot = NULL;
	conf.mimefile = NULL;
	conf.mimelist = NULL;
	conf.debug = -1;
	conf.user = DEFAULT_USER;
	conf.pidfile = DEFAULT_PIDFILE;

	while ((ch = getopt(argc, argv, "vdhc:")) != -1) {
		switch (ch) {
		case 'v':
			printf("%s/%s\n", getprogname(), MOHAWK_VERSION );
			exit(0);
		case 'c':
			parse_config(optarg);
			flag_config = 1;
			break;
		case 'd':
			conf.debug = 1;
			break;
		case '?':
		case 'h':
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;

	if (conf.debug < 0)
		conf.debug = 0;

	if (!flag_config) usage();

	openlog(getprogname(), LOG_NDELAY|LOG_PID, LOG_DAEMON);

	if (conf.mimefile != NULL) {
		if ((fd = open(conf.mimefile, O_RDONLY)) == -1)
			err(1, "open '%s'", conf.mimefile);

		conf.mimelist = properties_read(fd);
		if (conf.mimelist == NULL)
			errx(1, "An error occured while parsing mime list");

		close(fd);
	}

	/* If we're root and we're going to become another user, get the uid/gid
	** now.
	*/
	if (getuid() == 0) {
		pwd = getpwnam(conf.user);
		if (pwd == NULL) {
			syslog(LOG_CRIT, "unknown user - '%s'", conf.user);
			errx(1, "unknown user - '%s'", conf.user);
		}
		uid = pwd->pw_uid;
		gid = pwd->pw_gid;
	}

	if (SLIST_EMPTY(&conf.list)) {
		if (host("0.0.0.0", DEFAULT_HTTP_PORT) < 0 || host("::", DEFAULT_HTTP_PORT) < 0)
			errx(1, "Unable to listen on port '%s'", DEFAULT_HTTP_PORT);
	}

	SLIST_FOREACH(listener, &conf.list, entry) {
		if ((listener->fd = socket(listener->ai->ai_family, listener->ai->ai_socktype, listener->ai->ai_protocol)) == -1)
			//err(1, "socket()");
			continue;

		if (setsockopt(listener->fd, SOL_SOCKET, SO_REUSEADDR, (int[]){1}, sizeof(int)) < 0)
			err(1, "setsockopt()");

		if (bind(listener->fd, listener->ai->ai_addr, listener->ai->ai_addrlen) == -1)
			//err(1, "bind()");
			continue;

		if (listen(listener->fd, 1024) < 0)
			err(1, "listen()");
		nb_listen++;
	}

	if (nb_listen == 0)
		errx(1, "no listen. May be another mohawk process is runnig ?");

	if (!conf.debug) {
		if ((conf.pfh = pidfile_open(conf.pidfile, 0644, &otherpid)) == NULL) {
			if (errno == EEXIST)
				errx(EXIT_FAILURE, "Daemon already running, pid: %jd.", (intmax_t)otherpid);

			warn("Cannot open or create pidfile '%s'", conf.pidfile);
		}

		if (daemon(1, 0) < 0) {
			syslog(LOG_CRIT, "daemon - %m");
			pidfile_remove(conf.pfh);
			err(1, "daemon");
		}
		pidfile_write(conf.pfh);
	} else {
		/* Even if we don't daemonize, we still want to disown our parent
		** process.
		*/
		setsid();
	}

	/* Read zone info now, in case we chroot(). */
	tzset();

	/* If we're root, start becoming someone else. */
	if (getuid() == 0) {
		/* Set aux groups to null. */
		if (setgroups(0, (gid_t*) 0) < 0) {
			syslog(LOG_CRIT, "setgroups - %m");
			err(1, "setgroups");
		}
		/* Set primary group. */
		if (setgid(gid) < 0) {
			syslog(LOG_CRIT, "setgid - %m");
			err(1,  "setgid");
		}
		/* Try setting aux groups correctly - not critical if this fails. */
		if (initgroups(conf.user, gid) < 0) {
			syslog(LOG_ERR, "initgroups - %m");
			warn("initgroups");
		}
		/* Set login name. */
		setlogin(conf.user);
	}

	/* Chroot if requested. */
	if (conf.chroot) {
		if (chroot(conf.chroot) < 0) {
			syslog(LOG_CRIT, "chroot - %m");
			err(1, "chroot");
		}

		/* Always chdir to / after a chroot. */
		if (chdir("/") < 0) {
			syslog(LOG_CRIT, "chroot chdir - %m");
			err(1, "chroot chdir");
		}

	}

	/* If we're root, become someone else. */
	if (getuid() == 0) {
		/* Set uid. */
		if (setuid(uid) < 0) {
			syslog(LOG_CRIT, "setuid - %m");
			err(1, "setuid");
		}
		/* Check for unnecessary security exposure. */
		if (! conf.chroot) {
			syslog(LOG_WARNING,
				"started as root without requesting chroot(), warning only");
			warnx("started as root without requesting chroot(), warning only");
		}
	}

	/* Catch various signals. */
	sigset(SIGTERM, handle_sigterm);
	sigset(SIGINT, handle_sigterm);
	sigset(SIGUSR1, handle_sigterm);
	sigset(SIGHUP, handle_sighup);
	sigset(SIGCHLD, handle_sigchld);
	sigset(SIGPIPE, SIG_IGN);

	mohawk_serve();

	return(EXIT_SUCCESS);
}

static void
mohawk_serve(void)
{
	int kq;
	struct kevent ke;
	struct kevent *evlist = NULL;
	struct listener *listener;
	struct client *cl;
	int nev, i, r;
	int max_list_queues = 20;
	int nbevq = 0;

	if ((kq = kqueue()) == -1) {
		syslog(LOG_CRIT, "kqueue()");
		err(1, "kqueue");
	}

	SLIST_FOREACH(listener, &conf.list, entry) {
		EV_SET(&ke, listener->fd, EVFILT_READ, EV_ADD, 0, 0, NULL);
		kevent(kq, &ke, 1, NULL, 0, NULL);
		nbevq++;
	}

	for (;;) {
		if (evlist == NULL) {
			if ((evlist = malloc(max_list_queues * sizeof(struct kevent))) == NULL)
				errx(1, "Unable to allocate memory");
		}
		else if (nbevq > max_list_queues) {
			max_list_queues += 20;
			free(evlist);
			if ((evlist = malloc(max_list_queues * sizeof(struct kevent))) == NULL)
				errx(1, "Unable to allocate memory");
		}

		nev = kevent(kq, NULL, 0, evlist, max_list_queues, NULL);
		for (i = 0; i < nev; i++) {
			if (evlist[i].udata == NULL) {
				/* this is a listener */
				listener = get_listener_event(evlist[i].ident);
				char buf[NI_MAXHOST];
				char service[NI_MAXSERV];

				getnameinfo(listener->ai->ai_addr, listener->ai->ai_addrlen,
						buf, sizeof(buf),
						service, sizeof(service),
						NI_NUMERICHOST|NI_NUMERICSERV);
				if (conf.debug)
					fprintf(stderr, "Connect on %s[:%s]\n", buf, service);
				if ((cl = client_new(evlist[i].ident, service)) == NULL)
					continue;

				EV_SET(&ke, cl->fd, EVFILT_READ, EV_ADD, 0, 0, cl);
				kevent(kq, &ke, 1, NULL, 0, NULL);
				nbevq += 5;
				continue;
			} else {
				/* This is a client */
				cl = (struct client *)evlist[i].udata;

				if (evlist[i].flags & (EV_ERROR | EV_EOF) && evlist[i].filter != EVFILT_PROC) {
					client_free(cl);
					nbevq -= 5;
					continue;
				}

				if (evlist[i].filter == EVFILT_READ) {
					r = read_client(cl, evlist[i].ident, evlist[i].data);

					if (r == -1) {
						switch (cl->state) {
							case CLIENT_READ:
							case CGI_DONE:
								client_free(cl);
								nbevq -= 5;
								break;

							case CGI_HEADERS:
								send_error(cl, 500, "Internal Error", "", "Something went wrong when reading ouput of cgi");
								EV_SET(&ke, cl->fd, EVFILT_READ, EV_DELETE, 0, 0, 0);
								kevent(kq, &ke, 1, NULL, 0, NULL);
								EV_SET(&ke, cl->fd, EVFILT_WRITE, EV_ADD, 0, 0, cl);
								kevent(kq, &ke, 1, NULL, 0, NULL);
								cl->state = CGI_DONE;
								break;

							case CGI_BODY:
								cl->state = CGI_DONE;
								break;
						}
					} else if (r == 1) {
						switch (cl->state) {
							case CLIENT_READ:
								handle_request(cl);
								if (cl->state == CGI_HEADERS) {
									EV_SET(&ke, cl->cgi_out_fd, EVFILT_READ, EV_ADD, 0, 0, cl);
									kevent(kq, &ke, 1, NULL, 0, NULL);
									EV_SET(&ke, cl->cgi_err_fd, EVFILT_READ, EV_ADD, 0, 0, cl);
									kevent(kq, &ke, 1, NULL, 0, NULL);
									EV_SET(&ke, cl->cgi_pid, EVFILT_PROC, (EV_ADD | EV_ENABLE), NOTE_EXIT, 0, cl);
									kevent(kq, &ke, 1, NULL, 0, NULL);
								} else {
									EV_SET(&ke, cl->fd, EVFILT_WRITE, EV_ADD, 0, 0, cl);
									kevent(kq, &ke, 1, NULL, 0, NULL);
									EV_SET(&ke, cl->fd, EVFILT_READ, EV_DELETE, 0, 0, 0);
									kevent(kq, &ke, 1, NULL, 0, NULL);
								}
								break;
							case CGI_HEADERS:
								parse_cgi_headers(cl);
								if (cl->method == METHOD_HEAD) {
									EV_SET(&ke, cl->fd, EVFILT_WRITE, EV_ADD, 0, 0, cl);
									kevent(kq, &ke, 1, NULL, 0, NULL);
								} else {
									EV_SET(&ke, cl->fd, EVFILT_WRITE, EV_ADD, 0, 0, cl);
									kevent(kq, &ke, 1, NULL, 0, NULL);
								}
								break;
							case CGI_BODY:
								cl->state = CGI_DONE;
								break;
						}
					}

					continue;
				}

				if (evlist[i].filter == EVFILT_WRITE) {
					if (cl->state != CLIENT_SENDFILE) {
						r = write_client_data(cl, evlist[i].ident);
						if (r == -1) {
							client_free(cl);
							nbevq -= 5;
						} else if (r == 1) {
							switch (cl->state) {
								case CGI_BODY:
									continue;
								default:
									client_free(cl);
									nbevq -= 5;
									break;
							}
						}
					} else {
						off_t nread;
						if (cl->offset == 0)  {
							struct iovec iovec[1];

							bzero(&cl->sf_hdtr, sizeof(cl->sf_hdtr));
							bzero(&iovec, sizeof(iovec));

							if (!sbuf_done(cl->response))
								sbuf_finish(cl->response);
							iovec[0].iov_base = sbuf_data(cl->response);
							iovec[0].iov_len = sbuf_len(cl->response);
							cl->sf_hdtr.headers = iovec;
							cl->sf_hdtr.hdr_cnt = 1;
							cl->sf_hdtr.trailers = NULL;
							cl->sf_hdtr.trl_cnt = 0;
						}

						r = sendfile(cl->sendfile_fd, cl->fd, cl->offset, 0, cl->offset ? NULL : &cl->sf_hdtr, &nread, 0);

						cl->offset += cl->offset ? nread : nread - sbuf_len(cl->response);
						if (r == 0) {
							close(cl->sendfile_fd);
							EV_SET(&ke, cl->fd, EVFILT_READ, EV_DELETE, 0, 0, cl);
							kevent(kq, &ke, 1, NULL, 0, NULL);
							EV_SET(&ke, cl->fd, EVFILT_WRITE, EV_DELETE, 0, 0, cl);
							kevent(kq, &ke, 1, NULL, 0, NULL);
							client_free(cl);
							nbevq -= 5;
						}
					}
					continue;
				}

				if (evlist[i].filter == EVFILT_PROC) {
					cl->state = CGI_DONE;
					EV_SET(&ke, cl->fd, EVFILT_WRITE, EV_ADD, 0, 0, cl);
					kevent(kq, &ke, 1, NULL, 0, NULL);
					continue;
				}
			}
		}
	}
}

static struct client *
client_new(int fd, const char *service)
{
	socklen_t sz;
	int flags, ret;
	struct client *cl;

	/* Accept the new connection. */
	if ((cl = malloc(sizeof(struct client))) == NULL)
		errx(1, "Unable to allocate memory");
	if (snprintf(cl->server_port, NI_MAXSERV, "%s", service) < 0)
		errx(1, "snprintf");

	cl->response = sbuf_new_auto();
	cl->header_len = 0;
	cl->cgi_header_len = 0;
	cl->offset = 0;
	cl->nwrite = 0;
	cl->script_name = NULL;
	cl->query_string = NULL;
	cl->remote_user = NULL;
	cl->user_agent = NULL;
	cl->path_info = NULL;
	cl->referer = NULL;
	cl->cookie = NULL;
	cl->content_type = NULL;
	cl->authorization = NULL;
	cl->real_remote_addr = NULL;
	cl->content_length = 0;
	cl->if_modified_since = -1;
	cl->protocol = NULL;
	cl->status = 200;
	cl->state = CLIENT_READ;

	sz = sizeof(cl->ss);
	cl->fd = accept(fd, (struct sockaddr *)&(cl->ss), &sz);

	ret = getnameinfo((struct sockaddr *)&(cl->ss), cl->ss.ss_len, cl->remote_addr, sizeof(cl->remote_addr), NULL, 0, NI_NUMERICHOST|NI_NUMERICSERV);
	if (ret != 0)
		warnx("client_new getnameinfo(): %s", gai_strerror(ret));
	else if (conf.debug)
		fprintf(stderr, "Connect from '%s'\n", cl->remote_addr);

	if (cl->fd < 0) {
		if (errno == EINTR || errno == EAGAIN || errno == EPROTO) {
			client_free(cl);
			return (NULL);
		}

		syslog(LOG_CRIT, "accept - %m");
		err(1, "accept");
	}

	/* set non blocking */
	if (-1 == (flags = fcntl(cl->fd, F_GETFL, 0)))
		flags = 0;

	fcntl(cl->fd, F_SETFL, flags | O_NONBLOCK);
	return (cl);
}

static int
read_client(struct client *cl, int fd, size_t len)
{
	if ((cl->state == CGI_HEADERS || cl->state == CGI_BODY) && fd == cl->fd)
		return pass_cgi_stdin(cl, fd, len);

	if (fd == cl->fd)
		return read_http_header(cl, fd, len);

	if (cl->state == CGI_HEADERS && fd == cl->cgi_out_fd)
		return read_cgi_header(cl, fd, len);
	else
		return read_cgi_data(cl, fd, len);
}

/*
 * return -1 something went wrong
 * return 0 everything ok but read unfinished
 * return 1 everything ok and read finished
 */
static int
write_client_data(struct client *cl, int fd)
{
	int r;

	if (! sbuf_done(cl->response)) {
		sbuf_finish(cl->response);
	}
	r = mohawk_write(fd, sbuf_data(cl->response) + cl->nwrite, sbuf_len(cl->response) - cl->nwrite);
	if (r < 0 && (errno == EINTR || errno == EAGAIN))
		return (-1);

	cl->nwrite += r;
	if (cl->nwrite == sbuf_len(cl->response))
		return (1);

	return (0);
}

static int
pass_cgi_stdin(struct client *cl, int fd, int len)
{
	int r;
	char *buf;
	if ((buf = malloc(len)) == NULL)
		errx(1, "Unable to allocate memory");

	r = mohawk_read(fd, buf, len);
	if (r <= 0) {
		free(buf);
		if (errno == EAGAIN)
			return (0);

		return (-1);
	}

	mohawk_write(cl->cgi_in_fd, buf, r);

	return (0);
}

/*
 * return -1 something went wrong
 * return 0 everything ok but read unfinished
 * return 1 everything ok and read finished
 */
static int
read_cgi_data(struct client *cl, int fd, long len)
{
	int r;
	char *buf;

	if ((buf = malloc(len)) == NULL)
		errx(1, "Unable to allocate memory");

	r = mohawk_read(fd, buf, len);
	if (r < 0 && (errno == EINTR || errno == EAGAIN)) {
		free(buf);
		return (0);
	}

	if (r == 0) {
		free(buf);
		return (-1);
	}

	sbuf_bcat(cl->response, buf, r);
	free(buf);

	return (0);
}

static int
read_cgi_header(struct client *cl, int fd, long len)
{
	int r;
	char c;
	size_t recv_len = 0;

	while ((r = read(fd, &c, 1)) != -1) {
		if (r == 0)
			return (0);

		if (cl->cgi_header_len + 1 > 2048)
			return (-2);

		cl->cgi_header[cl->cgi_header_len] = c;
		cl->cgi_header_len++;
		recv_len++;

		if (c == '\n') {
			/* in case newline is LF */
			if (cl->cgi_header_len >= 2) {
				if (cl->cgi_header[cl->cgi_header_len - 2] == '\n') {
					cl->cgi_header[cl->cgi_header_len - 2] = '\0';
					return (1);
				}
			}
			/* in case new line is CRLF REF3875 says both should accepted */
			if (cl->cgi_header_len >= 4) {
				if (strncmp(cl->cgi_header + cl->cgi_header_len - 4, "\r\n\r\n", 4) == 0) {
					cl->cgi_header[cl->cgi_header_len - 4] = '\0';
					return (1);
				}
			}
		}

		if (recv_len == len)
			return (0);
	}
	return (-1);
}

/*
 * retun:
 * -2: header too long
 * -1: something went wrong
 *  0: everything is ok but read isn't finished
 *  1: everything ok and read is finished
 */
static int
read_http_header(struct client *cl, int fd, long len)
{
	int r;
	char c;
	size_t recv_len = 0;

	while ((r = recv(fd, &c, 1, 0)) != -1) {
		if (r == 0)
			return (0);

		if (cl->header_len + 1 > 8096)
			return (-2);

		cl->raw_header[cl->header_len] = c;
		cl->header_len++;
		recv_len++;

		if (c == '\n') {
			if (cl->header_len >= 4) {
				if (strncmp(cl->raw_header + cl->header_len - 4, "\r\n\r\n", 4) == 0) {
					cl->raw_header[cl->header_len - 4] = '\0';
					return (1);
				}
			}
		}

		if (recv_len == len)
			return (0);
	}

	if (errno == EAGAIN)
		return (0);

	return (-1);
}

static struct listener *
get_listener_event(int fd)
{
	struct listener *listener;

	SLIST_FOREACH(listener, &conf.list, entry)
		if (listener->fd == fd)
			return (listener);

	return (NULL);
}

static void
client_free(struct client *cl)
{
	struct mohawk_list *no_log_pattern;
	char *remote_addr;

	if (cl == NULL)
		return;

	int do_log = 1;
// FMT_LOG host remote_addr remote_user "method url protocol" status size
#define FMT_LOG "%s %s %s \"%s %s %s\" %lu %lu"
#define FMT_LOG_CGI "%s %s %s \"%s %s?%s %s\" %lu %lu"

	TAILQ_FOREACH(no_log_pattern, &cl->vhost->no_log_patterns, entry) {
		if (fnmatch(no_log_pattern->item, cl->script_name, 0) != FNM_NOMATCH) {
			do_log = 0;
			break;
		}
	}
	if (do_log) {
		if (cl->vhost->x_forwarded_for && cl->real_remote_addr)
			remote_addr = cl->real_remote_addr;
		else
			remote_addr = cl->remote_addr;

		if (cl->script_name == cl->query_string)
			syslog(LOG_ERR, FMT_LOG, cl->host, remote_addr,
				cl->remote_user ? cl->remote_user : "-",
				get_method_str(cl->method), 
				cl->script_name, cl->protocol,
				(unsigned long)cl->status, (unsigned long)sbuf_len(cl->response));
		else
			syslog(LOG_ERR, FMT_LOG_CGI, cl->host, remote_addr, 
				cl->remote_user ? cl->remote_user : "-",
				get_method_str(cl->method),
				cl->script_name, cl->query_string, cl->protocol,
				(unsigned long)cl->status, (unsigned long)sbuf_len(cl->response));
	}

	sbuf_delete(cl->response);
	close(cl->fd);
	free(cl);
}

static void
usage(void)
{
	errx(1, "usage: %s -c configfile [-d] [-h] [-v]\n", getprogname());
}

static void
parse_cgi_headers(struct client *cl)
{
	char *cp;
	char *title;
	char timebuf[100];
	time_t now;
	const char *rfc1123_fmt = "%a, %d %b %Y %H:%M:%S GMT";

	now = time(NULL);
	strftime(timebuf, sizeof(timebuf), rfc1123_fmt, gmtime(&now));
	sbuf_printf(cl->response, "Date: %s\r\n", timebuf);

	if (strlen(cl->cgi_header) == 0)
		return;

	cl->status = 200;
	if ((cp = strstr(cl->cgi_header, "Status:")) != NULL) {
		cp += 7;
		cp += strspn(cp, " \t");
		cl->status = strtol(cp, NULL, 10);
	}

	if ((cp = strstr(cl->cgi_header, "Location:")) != NULL) {
		cl->status = 302;
	}

	switch(cl->status) {
		case 200: title = "OK"; break;
		case 302: title = "Found"; break;
		case 304: title = "Not Modified"; break;
		case 400: title = "Bad Request"; break;
		case 401: title = "Unauthorized"; break;
		case 403: title = "Forbidden"; break;
		case 404: title = "Not Found"; break;
		case 408: title = "Request Timeout"; break;
		case 500: title = "Internal Error"; break;
		case 501: title = "Not Implemented"; break;
		case 503: title = "Service Temporarily Overloaded"; break;
		default: title = "Something"; break;
	}

	sbuf_clear(cl->response);
	sbuf_printf(cl->response, "HTTP/1.0 %d %s\r\n"
			"Server: %s/%s\r\n"
			"Date: %s\r\n"
			"Connection: close\r\n"
			, cl->status, title, getprogname(), MOHAWK_VERSION, timebuf);
	sbuf_printf(cl->response, "%s\r\n\r\n", cl->cgi_header);

	cl->state = CGI_BODY;
}

/* This runs in a child process, and exits when done, so cleanup is
** not needed.
*/
static void
handle_request(struct client *cl)
{
	char *method_str;
	char filepath[MAXPATHLEN];
	char *file;
	char *buf;
	char *line;
	struct stat sb;
	int r, file_len;
	char location[MAXPATHLEN];
	struct mohawk_hash *cgi_map;
	struct mohawk_list *restrict_from;
	struct vhost *vh;
	int restrict_ok = 0;

	method_str = cl->raw_header;

	if ((cl->script_name = strpbrk(method_str, " \t\r\n")) == NULL) {
		send_error(cl, 400, "Bad Request", "", "Can't parse request.");
		return;
	}

	cl->script_name[0] = '\0';
	cl->script_name++;

	if (strcmp(method_str, "GET") == 0)
		cl->method = METHOD_GET;
	else if (strcmp(method_str, "POST") == 0)
		cl->method = METHOD_POST;
	else if (strcmp(method_str, "HEAD") == 0)
		cl->method = METHOD_HEAD;
	else
		cl->method = METHOD_UNKNOWN;

	cl->script_name += strspn(cl->script_name, " \t\r\n");
	cl->protocol += strspn(cl->script_name, " \t\r\n");

	if ((cl->query_string = strchr(cl->script_name, '?')) != NULL)
		*cl->query_string++ = '\0';
	else
		cl->query_string = cl->script_name;

	if ((cl->protocol = strpbrk(cl->query_string, " \t\r\n")) == NULL) {
		send_error(cl, 400, "Bad Request", "", "Can't parse request.");
		return;
	}

	*cl->protocol++ = '\0';
	buf = cl->protocol;
	while ((line = strstr(buf, "\r\n")) != NULL) {
		line[0] = '\0';
		parse_hdrs_line(cl, buf);
		buf = line + 2;
	}
	parse_hdrs_line(cl, buf);

	SLIST_FOREACH(vh, &conf.vhosts, entry) {
		if (strcmp(vh->host, cl->host) == 0)
			break;
		if (fnmatch(vh->host, cl->host, 0) != FNM_NOMATCH)
			break;
	}
	/* client keep a ref to the vhost */
	cl->vhost = vh;

	if (conf.debug) {
		fprintf(stderr, "Debug: vhost '%s'\n", vh->host);
		fprintf(stderr, "Debug: script_name: '%s'\n", cl->script_name);
		if (cl->query_string != cl->script_name)
			fprintf(stderr, "Debug: query_string: '%s'\n", cl->query_string);
	}

	/* Simple restrict from: no regex, no netmask ... */
	if (! TAILQ_EMPTY(&vh->restrict_froms)) {
		TAILQ_FOREACH(restrict_from, &vh->restrict_froms, entry) {
			if (strncmp(restrict_from->item, cl->remote_addr, strlen(restrict_from->item)) == 0) {
				restrict_ok = 1;
				break;
			}
		}
		if (! restrict_ok) {
			send_error(cl, 403, "Access denied", "", "Access denied");
			if (conf.debug)
				fprintf(stderr, "Debug: '%s': access denied\n",  cl->remote_addr);
			return;
		}
	}

	/* script_name match cgi-map ? */
	TAILQ_FOREACH(cgi_map, &vh->cgi_map, entry) {
		if (fnmatch(cgi_map->key, cl->script_name, 0) != FNM_NOMATCH) {
			if (conf.debug)
				fprintf(stderr, "Debug: '%s' match '%s'\n", cl->script_name, cgi_map->key);
			/* Detemine the file path */
			if (strcmp(vh->host, "*") == 0)
				snprintf(filepath, MAXPATHLEN, "%s%s",  vh->rootdir, cgi_map->value);
			else
				snprintf(filepath, MAXPATHLEN, "%s/%s%s",  vh->rootdir, cl->host, cgi_map->value);

			do_cgi(cl, filepath);
			return;
		}
	}

	if (vh->status_url && strcmp(cl->script_name, vh->status_url) == 0) {
		do_status(cl);
		return;
	}

	/* Detemine the file path */
	if (strcmp(vh->host, "*") == 0)
		snprintf(filepath, MAXPATHLEN, "%s%s",  vh->rootdir, cl->script_name);
	else
		snprintf(filepath, MAXPATHLEN, "%s/%s%s",  vh->rootdir, cl->host, cl->script_name);

	/* Set up the timeout for writing. */
/*	(void) sigset(SIGALRM, handle_write_timeout);
	(void) alarm(WRITE_TIMEOUT);*/

	r = stat(filepath, &sb);

	if (r < 0) {
		send_error(cl, 404, "Not Found", "", "File not found.");
		return;
	}

	file_len = strlen(filepath);
	if (! S_ISDIR(sb.st_mode)) {
		do_file(cl, filepath, &sb);
	} else {
		/* The filename is a directory.  Is it missing the trailing slash? */
		if (filepath[file_len - 1] != '/' ) {
			if (cl->query_string != NULL && cl->script_name != cl->query_string)
				snprintf(location, sizeof(location), "Location: %s/?%s", cl->script_name,
					cl->query_string);
			else
				snprintf(location, sizeof(location), "Location: %s/", cl->script_name);
			send_error(cl, 302, "Found", location, "Directories must end with a slash.");
			return;
		}

		if (get_index(filepath, &file) == -1) {
			do_dir(cl, filepath, &sb);
			return;
		} else {
			stat(file, &sb);
			do_file(cl, file, &sb);
			free(file);
		}
	}
}

static int
get_index(char *dir, char **file)
{
	struct stat sb;
	struct mohawk_list *index_name;

	if ((*file = malloc(MAXPATHLEN)) == NULL)
		errx(1, "Unable to allocate memory");

	TAILQ_FOREACH(index_name, &vh->index_names, entry) {
		snprintf(*file, MAXPATHLEN, "%s%s", dir, index_name->item);
		if (stat(*file, &sb) >= 0)
			return (0);
	}

	free(*file);
	return (-1);
}

static void
do_status(struct client *cl)
{
	struct rusage ru;
	char buf[500];
	int rc;

	rc = getrusage(0, &ru);
	if (rc == -1) {
		syslog(LOG_INFO, "getrusage()");
		send_error(cl, 500, "Internal Error", "", "getrusage() return -1");
		return;
	}

#define STATUS_HEADER "stime\tutime\tmaxrss\tixrss\tidrss\tisrss\tminflt\tmajflt\tnswap\tinblock\toublock\tmsgsnd\tmsgrcv\tnsignals\tnvcsw\tnivcsw"
#if defined(__arm__) 
#define STATUS_FMT "<pre>\n%s\n%lld.%06ld\t%lld.%06ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld</pre>\n"
#elif defined(__LP64__)
#define STATUS_FMT "<pre>\n%s\n%ld.%06ld\t%ld.%06ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld</pre>\n"
#else
#define STATUS_FMT "<pre>\n%s\n%d.%06ld\t%d.%06ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld\t%ld</pre>\n"
#endif

	snprintf(buf, 500, STATUS_FMT, STATUS_HEADER,
		ru.ru_stime.tv_sec, ru.ru_stime.tv_usec,
		ru.ru_utime.tv_sec, ru.ru_utime.tv_usec,
		ru.ru_maxrss, ru.ru_ixrss, ru.ru_idrss, ru.ru_isrss, ru.ru_minflt, ru.ru_majflt,
		ru.ru_nswap,ru.ru_inblock, ru.ru_oublock, ru.ru_msgsnd, ru.ru_msgrcv,
		ru.ru_nsignals, ru.ru_nvcsw, ru.ru_nivcsw
	);
	send_error(cl, 200, "Status", "", buf);
}

static void
do_file(struct client *cl, char *filepath, struct stat *sb)
{
	const char *mime_type = NULL;
	char fixed_mime_type[500];
	char *filename;
	char *ext;
	char dirname[MAXPATHLEN];
	int fd;
	struct mohawk_list *cgi_pattern;

	if (conf.debug)
		fprintf(stderr, "Debug: start do_file '%s'\n", filepath);

	/* Check authorization for this directory. */
	snprintf(dirname, MAXPATHLEN, "%s", filepath);

	filename = strrchr(dirname, '/');
	filename[0] = '\0';
	filename++;

	if (auth_check(cl, dirname) != 0)
		return;

	/* Check if the filename is the AUTH_FILE itself - that's verboten. */
	if (strcmp(filename, AUTH_FILE) == 0) {
		syslog(LOG_NOTICE, "%.80s URL \"%.80s\" tried to retrieve an auth file", cl->remote_addr, cl->script_name);
		send_error(cl, 403, "Forbidden", "", "File is protected.");
		return;
	}

	/* Is it CGI? */
	TAILQ_FOREACH(cgi_pattern, &vh->cgi_patterns, entry) {
		if (fnmatch(cgi_pattern->item, filename, FNM_CASEFOLD) == 0) {
			do_cgi(cl, filepath);
			return;
		}
	}

	if ((fd = open(filepath, O_RDONLY)) < 0) {
		syslog(LOG_INFO, "%.80s File \"%.80s\" is protected", cl->remote_addr, cl->script_name);
		send_error(cl, 403, "Forbidden", "", "File is protected.");
		return;
	}

	ext = strrchr(filepath, '.');
	if (ext != NULL) {
		ext++;
		if (ext != '\0')
			mime_type = property_find(conf.mimelist, ext);
	}

	if (mime_type == NULL)
		mime_type = "application/octet-stream";

	snprintf(fixed_mime_type, sizeof(fixed_mime_type), mime_type, vh->charset);

	if (cl->if_modified_since != -1 &&
		cl->if_modified_since >= sb->st_mtime) {
		add_headers(cl, 304, "Not Modified", "", NULL, fixed_mime_type, (off_t) -1, sb->st_mtime);
		return;
	}
	add_headers(cl, 200, "Ok", "", NULL, fixed_mime_type, sb->st_size, sb->st_mtime);
	if (cl->method == METHOD_HEAD)
		return;

	if (sb->st_size > 0) {
		cl->state = CLIENT_SENDFILE;
		cl->offset = 0;
		cl->sendfile_fd = fd;
	}
}


static void
do_dir(struct client *cl, char *path, struct stat *sb)
{
	struct sbuf *buf;
	int n, i;
	struct dirent **dl;

	if (conf.debug)
		fprintf(stderr, "Debug: start do_dir '%s'\n", path);

	buf = sbuf_new_auto();
	/* Check authorization for this directory. */
	if (auth_check(cl, path) != 0)
		return;

	n = scandir(path, &dl, NULL, alphasort);
	if (n < 0) {
		syslog(LOG_INFO, "%.80s Directory \"%.80s\" is protected", cl->remote_addr, cl->script_name);
		send_error(cl, 403, "Forbidden", "", "Directory is protected.");
		return;
	}

	sbuf_printf(buf, "\
<HTML>\n\
<HEAD><TITLE>Index of %s</TITLE>\n", cl->script_name);
       if (vh->dirlist_css_url) {
               sbuf_printf(buf, "\
<link href='%s' rel='stylesheet' type='text/css' />\n", vh->dirlist_css_url);
       }
       sbuf_printf(buf, "\
<BODY>\n\
<H2>Index of %s</H2>\n\
<div class='list'>\n\
<table summary='Directory Listing' cellpadding='0' cellspacing='0'>\n\
<thead><tr><th class='n'>Name</th><th class='m'>Last Modified</th><th class='s'>Size</th><th class='t'>Type</th></tr></thead>\n\
<tbody>\n", cl->script_name);

	for (i = 0; i < n; ++i)
		file_details(buf, path, dl[i]->d_name, cl->script_name);

	sbuf_printf(buf, "\
</tbody>\n\
</table>\n\
</div>\n\
<div class='foot'>%s/%s</div>\n\
</BODY>\n\
</HTML>\n",
	getprogname(), MOHAWK_VERSION);

	add_headers(cl, 200, "Ok", "", "", "text/html; charset=%s", sbuf_len(buf), sb->st_mtime);
	if (cl->method != METHOD_HEAD) {
		sbuf_finish(buf);
		sbuf_bcat(cl->response, sbuf_data(buf), sbuf_len(buf));
	}
	sbuf_delete(buf);
}

static void
file_details(struct sbuf *buf, const char *dir, const char *name, const char *script_name)
{
	struct stat sb;
	char f_time[20];
	static char encname[MAXPATHLEN];
	char path[MAXPATHLEN];
	char sz[6];
	const char *ext;
	const char *mime = NULL;

	snprintf(path, MAXPATHLEN, "%s/%s", dir, name);

	if (lstat(path, &sb) < 0) {
		sbuf_bcat(buf, "???", 3);
		return;
	}

	strftime(f_time, sizeof(f_time), "%d-%b-%Y %H:%M", localtime(&sb.st_mtime));
	snprintf(path, MAXPATHLEN, "%s%s", script_name, name);
	strencode(encname, MAXPATHLEN, path);
	humanize_number(sz, 6, sb.st_size, "", HN_AUTOSCALE, HN_B | HN_DECIMAL | HN_NOSPACE);

	/* no need to have . listed */
	if (strcmp(name, ".") == 0)
		return;

	if ((ext = strrchr(name, '.')) != NULL) {
		ext++;
		mime = property_find(conf.mimelist, ext);
	}

	if (mime == NULL)
		mime = "File";

	if (sb.st_mode & S_IFDIR)
		sbuf_printf(buf, "<tr><td class='n'><a href='%s/'>%s/</a></td><td class='m'>%s</td><td class='s'>-</td><td class='t'>Directory</td></tr>\n", encname, name, f_time);
	else
		sbuf_printf(buf, "<tr><td class='n'><a href='%s'>%s</a></td><td class='m'>%s</td><td class='s'>%s</td><td class='t'>%s</td></tr>\n", encname, name, f_time, sz, mime);
}

/* Copies and encodes a string. */
static void
strencode(char *to, size_t tosize, const char *from)
{
	int tolen;

	for (tolen = 0; *from != '\0' && tolen + 4 < tosize; ++from) {
		if (isalnum(*from) || strchr("/_.-~", *from) != (char*) 0) {
			*to = *from;
			++to;
			++tolen;
		} else {
			(void) sprintf(to, "%%%02x", (int) *from & 0xff);
			to += 3;
			tolen += 3;
		}
	}
	*to = '\0';
}

static pid_t
exec_cgi(struct client *cl, const char *command, int *infp, int *outfp, int *errfp)
{
	int p_stdin[2], p_stdout[2], p_stderr[3];
	pid_t pid;
	char** envp;
	char *argp[2];

	if (pipe(p_stdin) != 0 || pipe(p_stdout) != 0 || pipe(p_stderr) != 0)
		return -1;

	pid = fork();

	if (pid < 0)
		return pid;
	else if (pid == 0) {
		close(p_stdin[CGI_WRITE]);
		close(p_stdout[CGI_READ]);
		close(p_stderr[CGI_READ]);

		dup2(p_stdin[CGI_READ], CGI_READ);
		dup2(p_stdout[CGI_WRITE], CGI_WRITE);
		dup2(p_stderr[CGI_WRITE], CGI_ERR);

		nice(CGI_NICE);
		/* Make the environment vector. */
		envp = make_envp(cl);

		argp[0] = strrchr(command,'/');
		argp[1] = NULL;
		execve(command, argp, envp);
	}

	if (infp == NULL)
		close(p_stdin[CGI_WRITE]);
	else
		*infp = p_stdin[CGI_WRITE];

	if (outfp == NULL)
		close(p_stdout[CGI_READ]);
	else
		*outfp = p_stdout[CGI_READ];

	if (errfp == NULL)
		close(p_stderr[CGI_READ]);
	else
		*errfp = p_stderr[CGI_READ];

	return pid;
}

static void
do_cgi(struct client *cl, char *filepath)
{
	if (conf.debug)
		fprintf(stderr, "Debug: do_cgi '%s'\n", filepath);

	struct stat sb;
	if (stat(filepath, &sb) < 0) {
		send_error(cl, 500, "Internal Error", "", "No cgi found.");
		return;
	}
	cl->state = CGI_HEADERS;
	cl->cgi_pid = exec_cgi(cl, filepath, &cl->cgi_in_fd, &cl->cgi_out_fd, &cl->cgi_err_fd);
	return;
}

/* Set up CGI environment variables. Be real careful here to avoid
** letting malicious clients overrun a buffer.  We don't have
** to worry about freeing stuff since we're a sub-process.
*/
static char**
make_envp(struct client *cl)
{
#define MAX_ENV 50
	static char *envp[MAX_ENV + 1];
	int envn = 0;
	struct sbuf *buf = sbuf_new_auto();
	struct mohawk_hash *cgi_env;

	/* from http://tools.ietf.org/html/rfc3875 */
	if (cl->authorization != NULL)
		envp[envn++] = build_env(buf, "AUTH_TYPE=%s", "Basic");
	if (cl->content_length > 0)
		envp[envn++] = build_env(buf, "CONTENT_LENGTH=%lu", (unsigned long) cl->content_length);
	if (cl->content_type != NULL)
		envp[envn++] = build_env(buf, "CONTENT_TYPE=%s", cl->content_type);
	envp[envn++] = "GATEWAY_INTERFACE=CGI/1.1";
	if (cl->path_info != NULL) {
		envp[envn++] = build_env(buf, "PATH_INFO=/%s", cl->path_info);
		envp[envn++] = build_env(buf, "PATH_TRANSLATED=/%s", cl->path_info);
	}
	if (cl->query_string != NULL)
		envp[envn++] = build_env(buf, "QUERY_STRING=%s", cl->query_string);
	envp[envn++] = build_env(buf, "REMOTE_ADDR=%s", cl->remote_addr);

    //REMOTE_HOST
    //REMOTE_IDENT

    if (cl->remote_user != NULL)
        envp[envn++] = build_env(buf, "REMOTE_USER=%s", cl->remote_user);
	envp[envn++] = build_env(buf, "REQUEST_METHOD=%s", get_method_str(cl->method));
	envp[envn++] = build_env(buf, "SCRIPT_NAME=%s", cl->script_name);

    //SERVER_NAME
	envp[envn++] = build_env(buf, "SERVER_PORT=%s", cl->server_port);

	envp[envn++] = "SERVER_PROTOCOL=HTTP/1.0";
	envp[envn++] = build_env(buf, "SERVER_SOFTWARE=mohawk/%s", MOHAWK_VERSION);

	/* mohawk env */
	envp[envn++] = build_env(buf, "PATH=%s", CGI_PATH);
	envp[envn++] = build_env(buf, "LD_LIBRARY_PATH=%s", CGI_LD_LIBRARY_PATH);
	envp[envn++] = build_env(buf, "MOHAWK_VERSION=%s", MOHAWK_VERSION);

	/* client env */
	if (cl->referer != NULL)
		envp[envn++] = build_env(buf, "HTTP_REFERER=%s", cl->referer);
	if (cl->user_agent != NULL)
		envp[envn++] = build_env(buf, "HTTP_USER_AGENT=%s", cl->user_agent);
	if (cl->cookie != NULL)
		envp[envn++] = build_env(buf, "HTTP_COOKIE=%s", cl->cookie);
	if (cl->host != NULL)
		envp[envn++] = build_env(buf, "HTTP_HOST=%s", cl->host);

	/* system env */
	if (getenv("TZ") != (char*) 0)
		envp[envn++] = build_env(buf, "TZ=%s", getenv("TZ"));

	TAILQ_FOREACH(cgi_env, &vh->cgi_env, entry) {
		envp[envn++] = build_env(buf, "%s=%s", cgi_env->key, cgi_env->value);
		if (envn >= MAX_ENV) break;
	}
	sbuf_delete(buf);
	envp[envn] = (char*) 0;
	return envp;
}

static char*
build_env(struct sbuf *buf, char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	sbuf_clear(buf);
	sbuf_vprintf(buf, fmt, ap);
	va_end(ap);
	sbuf_finish(buf);
	return strdup(sbuf_data(buf));
}

/*
 * 0: auth ok
 * 1: auth ko
 */
static int
auth_check(struct client *cl, char *dirname)
{
	char authpath[MAXPATHLEN];
	struct stat sb;
	char authinfo[500];
	char *authpass;
	char *colon;
	char *line = NULL;
	size_t linecap = 0;
	ssize_t l;
	FILE* fp;
	char *cryp;

	/* Construct auth filename. */
	if (dirname[strlen(dirname) - 1] == '/')
		snprintf(authpath, sizeof(authpath), "%s%s", dirname, AUTH_FILE);
	else
		snprintf(authpath, sizeof(authpath), "%s/%s", dirname, AUTH_FILE);

	/* Does this directory have an auth file? */
	if (stat(authpath, &sb) < 0)
		/* Nope, let the request go through. */
		return (0);

	/* Does this request contain authorization info? */
	if (cl->authorization == NULL) {
		/* Nope, return a 401 Unauthorized. */
		send_authenticate(cl, dirname);
		return (1);
	}

	/* Basic authorization info? */
	if (strncmp(cl->authorization, "Basic ", 6) != 0) {
		send_authenticate(cl, dirname);
		return (1);
	}

	/* Decode it. */
	l = b64_pton(cl->authorization + 6, (u_char *)authinfo, sizeof(authinfo));
	authinfo[l] = '\0';
	/* Split into user and password. */
	authpass = strchr(authinfo, ':');
	if (authpass == NULL) {
		/* No colon?  Bogus auth info. */
		send_authenticate(cl, dirname);
		return (1);
	}
	*authpass++ = '\0';
	/* If there are more fields, cut them off. */
	colon = strchr(authpass, ':');
	if (colon != NULL)
		*colon = '\0';

	/* Open the password file. */
	if ((fp = fopen(authpath, "r")) == NULL) {
		/* The file exists but we can't open it?  Disallow access. */
		syslog(LOG_ERR, "%.80s auth file %.80s could not be opened - %m",
			cl->remote_addr, authpath);
		send_error(cl, 403, "Forbidden", "", "File is protected.");
		return (1);
	}

	/* Read it. */
	while ((l = getline(&line, &linecap, fp)) > 0) {
		/* Nuke newline. */
		l = strlen(line);
		if (line[l - 1] == '\n')
			line[l - 1] = '\0';

		/* Split into user and encrypted password. */
		cryp = strchr(line, ':');
		if (cryp == NULL)
			continue;
		*cryp++ = '\0';
		/* Is this the right user? */
		if (strcmp(line, authinfo) == 0) {
			/* Yes. */
			fclose(fp);
			/* So is the password right? */
			if (strcmp(crypt(authpass, cryp), cryp) == 0) {
				/* Ok! */
				cl->remote_user = line;
				return (0);
			} else {
				/* No. */
				send_authenticate(cl, dirname);
				return (1);
			}
		}
	}

	/* Didn't find that user.  Access denied. */
	fclose(fp);
	send_authenticate(cl, dirname);
	return (1);
}

static void
send_authenticate(struct client *cl, char *realm)
{
	char header[10000];

	snprintf(header, sizeof(header), "WWW-Authenticate: Basic realm=\"%s\"", realm);
	send_error(cl, 401, "Unauthorized", header, "Authorization required.");
}

static void
send_error(struct client *cl, int s, char *title, char *extra_header, char *text)
{
	add_headers(cl, s, title, extra_header, "", "text/html; charset=%s", (off_t) -1, (time_t) -1);

	sbuf_printf(cl->response,
"<HTML>\n"
"<HEAD><TITLE>%d %s</TITLE></HEAD>"
"<BODY>\n"
"<H4>%d %s</H4>\n"
"%s"
"<HR>\n"
"%s/%s\n"
"</BODY>\n"
"</HTML>\n" ,
		s, title, s, title, text,
	getprogname(), MOHAWK_VERSION);

	return;
}

static void
add_headers(struct client *cl, int s, char *title, char *extra_header, char *me, char *mt, off_t b, time_t mod)
{
	time_t now, expires;
	char timebuf[100];
	int s100;
	const char *rfc1123_fmt = "%a, %d %b %Y %H:%M:%S GMT";

	cl->status = s;
	sbuf_printf(cl->response, "%s %d %s\r\n", cl->protocol, cl->status, title);

	sbuf_printf(cl->response, "Server: %s/%s\r\n", getprogname(), MOHAWK_VERSION);

	now = time((time_t*) 0);
	strftime(timebuf, sizeof(timebuf), rfc1123_fmt, gmtime(&now));
	sbuf_printf(cl->response, "Date: %s\r\n", timebuf);

	s100 = cl->status / 100;
	if (s100 != 2 && s100 != 3)
		sbuf_printf(cl->response, "Cache-Control: no-cache,no-store\r\n");

	if (extra_header != NULL && extra_header[0] != '\0')
		sbuf_printf(cl->response, "%s\015\012", extra_header);

	if (me != NULL && me[0] != '\0')
		sbuf_printf(cl->response, "Content-Encoding: %s\r\n", me);

	if (mt != NULL && mt[0] != '\0') {
		if (strchr(mt, '%')) {
			sbuf_printf(cl->response, "Content-Type: ");
			sbuf_printf(cl->response, mt, cl->vhost->charset);
			sbuf_printf(cl->response, "\r\n");
		}
		else {
			sbuf_printf(cl->response, "Content-Type: %s\r\n", mt);
		}
	}

	if (b >= 0)
		sbuf_printf(cl->response, "Content-Length: %lld\r\n", (long long int) b);

	if (s100 == 2 && cl->vhost->maxage >= 0) {
		expires = now + cl->vhost->maxage;
		strftime(timebuf, sizeof(timebuf), rfc1123_fmt, gmtime(&expires));
		sbuf_printf(cl->response, "Cache-Control: max-age=%d\r\nExpires: %s\r\n", cl->vhost->maxage, timebuf);
	}

	if (mod != -1) {
		strftime(timebuf, sizeof(timebuf), rfc1123_fmt, gmtime(&mod));
		sbuf_printf(cl->response, "Last-Modified: %s\r\n", timebuf);
	}

	sbuf_printf(cl->response, "Connection: close\r\n\r\n");
}

static ssize_t
mohawk_read(int fd, char *buf, size_t size)
{
	return read(fd, buf, size);
}

static ssize_t
mohawk_write(int fd, char *buf, size_t size)
{
	return write(fd, buf, size);
}

static char*
get_method_str(int m)
{
	switch (m) {
		case METHOD_GET: return "GET";
		case METHOD_HEAD: return "HEAD";
		case METHOD_POST: return "POST";
		default: return "UNKNOWN";
	}
}

static void
handle_sigterm(int sig)
{
	/* Don't need to set up the handler again, since it's a one-shot. */

	syslog(LOG_NOTICE, "exiting due to signal %d", sig);
	closelog();

	if (conf.mimelist != NULL)
		properties_free(conf.mimelist);

	pidfile_remove(conf.pfh);
	errx(1, "exiting due to signal %d", sig);
}

/* SIGHUP says to re-open the log file. */
static void
handle_sighup(int sig)
{
	const int oerrno = errno;

	/* Just set a flag that we got the signal. */

	/* Restore previous errno. */
	errno = oerrno;
}


static void
handle_sigchld(int sig)
{
	const int oerrno = errno;
	pid_t pid;
	int status;

	/* Reap defunct children until there aren't any more. */
	for (;;) {
		pid = waitpid((pid_t) -1, &status, WNOHANG);

		if ((int) pid == 0)		/* none left */
			break;
		if ((int) pid < 0) {
			if (errno == EINTR || errno == EAGAIN)
				continue;
			/* ECHILD shouldn't happen with the WNOHANG option,
			** but with some kernels it does anyway.  Ignore it.
			*/
			if (errno != ECHILD) {
				syslog(LOG_ERR, "child wait - %m");
				warn("child wait");
			}
			break;
		}
	}

	/* Restore previous errno. */
	errno = oerrno;
}

/*static void
handle_read_timeout(int sig, struct client cl)
{
	syslog(LOG_INFO, "%.80s connection timed out reading", ntop(&cl->ss));
	send_error(cl, 
		408, "Request Timeout", "",
		"No request appeared within a reasonable time period.");
}*/

/*static void
handle_write_timeout(int sig)
{
	syslog(LOG_INFO, "%.80s connection timed out writing", ntop(&cl->ss));
	exit(1);
}*/

static void
parse_hdrs_line(struct client *cl, char *line)
{
	char *buf;
	char *cp;

	buf = strchr(line, ':');
	switch(buf - line + 1) {
		case 14:
			if (strncasecmp(line, "Authorization:", 14) == 0) {
				cp = line + 14;
				cp += strspn(cp, " \t");
				cl->authorization = cp;
			}
			break;
		case 15:
			if (strncasecmp(line, "Content-Length:", 15) == 0) {
				cp = &line[15];
				cp += strspn(cp, " \t");
				cl->content_length = atol(cp);
			}
			break;
		case 16:
			if (strncasecmp(line, "X-Forwarded-For:", 16) == 0) {
				cp = &line[16];
				cp += strspn(cp, " \t");
				cl->real_remote_addr = cp;
			}
			break;
		case 13:
			if (strncasecmp(line, "Content-Type:", 13) == 0) {
				cp = &line[13];
				cp += strspn(cp, " \t");
				cl->content_type = cp;
			}
			break;
		case 7:
			if (strncasecmp(line, "Cookie:", 7) == 0) {
				cp = &line[7];
				cp += strspn(cp, " \t");
				cl->cookie = cp;
			}
			break;
		case 5:
			if (strncasecmp(line, "Host:", 5) == 0) {
				char *pos;
				cp = &line[5];
				cp += strspn(cp, " \t");
				cl->host = cp;
                if ((pos = strchr(cl->host, ':')) != NULL) {
                    //host:port
                    *pos = '\0';
                }
			}
			break;
		case 18:
			if (strncasecmp(line, "If-Modified-Since:", 18) == 0) {
				cp = &line[18];
				cp += strspn(cp, " \t");
				cl->if_modified_since = parse_date(cp);
			}
			break;
		case 8:
			if (strncasecmp(line, "Referer:", 8) == 0) {
				cp = &line[8];
				cp += strspn(cp, " \t");
				cl->referer = cp;
			}
			break;
		case 11:
			if (strncasecmp(line, "User-Agent:", 11) == 0) {
				cp = &line[11];
				cp += strspn(cp, " \t");
				cl->user_agent = cp;
			}
			break;
	}
}

time_t
parse_date(const char *s)
{
	struct tm t;
	t.tm_isdst = 0;

	if (strptime(s, DATE_1123, &t) != NULL)
		return timegm(&t);

	if (strptime(s, DATE_1036, &t) != NULL)
		return timegm(&t);

	if (strptime(s, DATE_ANSI, &t) != NULL)
		return timegm(&t);

	return -1;
}

/* vim: set sw=4 sts=4 ts=4 : */
