/* mohawk - small HTTP server
**
** Copyright © 2010 by Baptiste Daroussin <bapt@FreeBSD.org>,
** Copyright © 2010 by Freddy Dissaux <freddy.dsx@free.fr>,
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
** OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
*/
%{
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <err.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <netdb.h>
#include <ifaddrs.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <pwd.h>
#include "mohawk.h"

#define WANT_FILENAME 0
#define WANT_DIRNAME  1

extern struct httpd conf;

struct vhost *vh = NULL;
struct vhost *vh_default = NULL;
struct mohawk_list *cgi_pattern = NULL;
struct mohawk_list *no_log_pattern = NULL;
struct mohawk_list *index_name = NULL;
struct mohawk_list *restrict_from = NULL;
struct mohawk_hash *cgi_env = NULL;
struct mohawk_hash *cgi_map = NULL;

int yyparse();

void
yyerror(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vprintf(fmt, ap);
	printf("\n");
	va_end(ap);
}

void
number_is_port(int port)
{
	if (port < 1 || port > 65535)
		errx(1, "'%d': not a valid port\n", port);
}

void
word_is_fs(const char *s, int want)
{
	struct stat sb;
	char buf[MAXPATHLEN];

	if (stat(s, &sb) != 0) {
		if (conf.chroot != NULL) {
			snprintf(buf, sizeof(buf), "%s%s", conf.chroot, s); 
			if (stat(buf, &sb) == 0) {
				if (S_ISDIR(sb.st_mode) && want == WANT_DIRNAME) { return; }
				if (S_ISREG(sb.st_mode) && want == WANT_FILENAME) { return; }
				errx(1, "'%s': not a %s\n", buf, want == WANT_DIRNAME ? "directory" : "file");
			}
		}
	}
	else {
		if (S_ISDIR(sb.st_mode) && want == WANT_DIRNAME) { return; }
		if (S_ISREG(sb.st_mode) && want == WANT_FILENAME) { return; }
	}
	errx(1, "'%s': not a %s\n", s, want == WANT_DIRNAME ? "directory" : "file");
}

void
word_is_login(const char *s)
{
	if (getpwnam(s) != NULL) { return; }
	errx(1, "'%s': not a valid user\n", s);
}

int yywrap()
{
	return 1;
} 

void
parse_config(const char *filename)
{
	int fd;

	if ((vh = malloc(sizeof(struct vhost))) == NULL)
		errx(1, "Unable to allocate memory");
	TAILQ_INIT(&vh->cgi_patterns);
	TAILQ_INIT(&vh->no_log_patterns);
	TAILQ_INIT(&vh->index_names);
	TAILQ_INIT(&vh->restrict_froms);
	TAILQ_INIT(&vh->cgi_env);
	TAILQ_INIT(&vh->cgi_map);
	vh->maxage = -1;
	vh->x_forwarded_for = 0;

	if ((cgi_pattern = malloc(sizeof(struct mohawk_list))) == NULL)
		errx(1, "Unable to allocate memory");

	if ((no_log_pattern = malloc(sizeof(struct mohawk_list))) == NULL)
		errx(1, "Unable to allocate memory");

    if ((index_name = malloc(sizeof(struct mohawk_list))) == NULL)
        errx(1, "Unable to allocate memory");

    if ((restrict_from = malloc(sizeof(struct mohawk_list))) == NULL)
        errx(1, "Unable to allocate memory");

	if ((cgi_map = malloc(sizeof(struct mohawk_hash))) == NULL)
		errx(1, "Unable to allocate memory");

	if ((cgi_env = malloc(sizeof(struct mohawk_hash))) == NULL)
		errx(1, "Unable to allocate memory");

	if ((fd = open(filename, O_RDONLY)) == -1)
		err(1, "%s", filename);

	if (close(STDIN_FILENO) == -1)
		err(1, "stdin");

	if (dup(fd) == -1)
		err(1, "%s", filename);

	yyparse();
	close(fd);

    if (vh_default == NULL)
        errx(1, "no 'default' vhost found");

}

int
host(const char *host, const char *port)
{
	struct listener *h;
	struct addrinfo *ai, *current_ai;
	struct addrinfo hints;
	int err;

	/* initialise addrinfo */
	memset(&hints, 0, sizeof hints);
	hints.ai_family = PF_UNSPEC;
	hints.ai_flags = AI_PASSIVE;
	hints.ai_socktype = SOCK_STREAM;

	if ((err = getaddrinfo(host, port, &hints, &ai)) != 0) {
		warnx("getaddrinfo(): %s", gai_strerror(err));
		return -1;
	}

	for (current_ai = ai; current_ai != NULL; current_ai = current_ai->ai_next) {
		if ((h = calloc(1, sizeof(struct listener))) == NULL)
			errx(1, "Unable to allocate memory");
		h->ai = current_ai;

		SLIST_INSERT_HEAD(&conf.list, h, entry);
	}

	return 0;
}
/*
-1: error
0: not a interface
1: ok
*/
int
interface(const char *s, const char *service)
{
	struct ifaddrs *ifap, *p;
	struct addrinfo *ai, *current_ai;
	struct addrinfo hints;
	struct listener         *h;
	char host[MAXHOSTNAMELEN];
	int family;
	int ret = 0;

	if (s == NULL) return 0;

	if (getifaddrs(&ifap) == -1)
		err(-1, "getifaddrs");

	for (p = ifap; p != NULL; p = p->ifa_next) {
		if (p->ifa_addr == NULL) continue;
		if (strcmp(s, p->ifa_name) != 0)
			continue;

		family = p->ifa_addr->sa_family;
		if (family != AF_INET && family != AF_INET6) continue;

		memset(&hints, 0, sizeof hints);
		hints.ai_family = PF_UNSPEC;
		hints.ai_flags = AI_PASSIVE;
		hints.ai_socktype = SOCK_STREAM;

		if ((ret = getnameinfo(p->ifa_addr, p->ifa_addr->sa_len, host, MAXHOSTNAMELEN, 0, 0, NI_NUMERICHOST)) != 0) {
			warnx("getnameinfo(): %s", gai_strerror(ret));
			ret = -1;
			break;
		}

		if ((ret = getaddrinfo(host, service, &hints, &ai)) != 0) {
			warnx("getaddrinfo(): %s '%s'", gai_strerror(ret), host);
			ret = -1;
			break;
		}

		ret = 1;
		for (current_ai = ai; current_ai != NULL; current_ai = current_ai->ai_next) {
			if ((h = calloc(1, sizeof(struct listener))) == NULL)
				errx(1, "Unable to allocate memory");
			h->ai = current_ai;

			SLIST_INSERT_HEAD(&conf.list, h, entry);
		}
	}

	freeifaddrs(ifap);

	return ret;
}

void
init_vhost(struct vhost *vh)
{

	struct mohawk_list *default_list;
	struct mohawk_list *list;

	struct mohawk_hash *default_hash;
	struct mohawk_hash *hash;

    if (vh_default == NULL)
        errx(1, "no 'default' vhost found");

    if (vh_default->rootdir == NULL)
        errx(1, "vhost 'default' without rootdir");

	vh->rootdir = vh_default->rootdir;
	vh->dirlist_css_url = vh_default->dirlist_css_url;
	vh->charset = vh_default->charset;
	vh->status_url = vh_default->status_url;
	vh->maxage = vh_default->maxage;
	vh->x_forwarded_for = vh_default->x_forwarded_for;

	TAILQ_FOREACH(default_list, &vh_default->cgi_patterns, entry) {
		if ((list = malloc(sizeof(struct mohawk_list))) == NULL)
			errx(1, "Unable to allocate memory");
		list->item = default_list->item;
		TAILQ_INSERT_TAIL(&vh->cgi_patterns, list, entry);
	}

	TAILQ_FOREACH(default_list, &vh_default->no_log_patterns, entry) {
		if ((list = malloc(sizeof(struct mohawk_list))) == NULL)
			errx(1, "Unable to allocate memory");
		list->item = default_list->item;
		TAILQ_INSERT_TAIL(&vh->no_log_patterns, list, entry);
	}

    TAILQ_FOREACH(default_list, &vh_default->index_names, entry) {
        if ((list = malloc(sizeof(struct mohawk_list))) == NULL)
            errx(1, "Unable to allocate memory");
        list->item = default_list->item;
        TAILQ_INSERT_TAIL(&vh->index_names, list, entry);
    }

    TAILQ_FOREACH(default_list, &vh_default->restrict_froms, entry) {
        if ((list = malloc(sizeof(struct mohawk_list))) == NULL)
            errx(1, "Unable to allocate memory");
        list->item = default_list->item;
        TAILQ_INSERT_TAIL(&vh->restrict_froms, list, entry);
    }

	TAILQ_FOREACH(default_hash, &vh_default->cgi_map, entry) {
		if ((hash = malloc(sizeof(struct mohawk_hash))) == NULL)
			errx(1, "Unable to allocate memory");
		hash->key = default_hash->key;
		hash->value = default_hash->value;
		TAILQ_INSERT_TAIL(&vh->cgi_map, hash, entry);
	}

	TAILQ_FOREACH(default_hash, &vh_default->cgi_env, entry) {
		if ((hash = malloc(sizeof(struct mohawk_hash))) == NULL)
			errx(1, "Unable to allocate memory");
		hash->key = default_hash->key;
		hash->value = default_hash->value;
		TAILQ_INSERT_TAIL(&vh->cgi_env, hash, entry);
	}
}

void
add_vhost(char *s) {
    vh->host = s;
    if (strcmp("default", s) == 0) {
        vh_default = vh;
        vh->host = "*";
    }

    SLIST_INSERT_HEAD(&conf.vhosts, vh, entry);
    if ((vh = malloc(sizeof(struct vhost))) == NULL)
        errx(1, "Unable to allocate memory");
    TAILQ_INIT(&vh->cgi_patterns);
    TAILQ_INIT(&vh->no_log_patterns);
    TAILQ_INIT(&vh->index_names);
    TAILQ_INIT(&vh->restrict_froms);
    TAILQ_INIT(&vh->cgi_map);
    TAILQ_INIT(&vh->cgi_env);
    init_vhost(vh);
}

void
add_cgi_env(char *key, char *value) {
	cgi_env->key = key;
	cgi_env->value = value;
	TAILQ_INSERT_TAIL(&vh->cgi_env, cgi_env, entry);
	if ((cgi_env = malloc(sizeof(struct mohawk_hash))) == NULL)
		errx(1, "Unable to allocate memory");
}


%}

%{
int yylex(void);
%}

/* commands */
%token SET LISTEN VHOST

/* keywords conf */
%token DEBUG CHROOT USER MIME PORT PIDFILE

/* keywords vhost */
%token ROOTDIR CGI_PATTERN DIRLIST_CSS_URL CHARSET STATUS_URL 
%token CGI_ENV OBRACE EBRACE CGI_MAP INDEX_NAME RESET NO_LOG
%token RESTRICT_FROM MAXAGE X_FORWARDED_FOR

/* values */
%token DEFAULT WORD NUMBER WORDS STATE

%union 
{
	int number;
	char *string;
}

%token <number> STATE
%token <number> NUMBER
%token <string> WORD
%token <string> WORDS

%%
commands: /* empty */
        | commands command
        ;

command:
       set
       |
       reset
       |
       listen
       |
       vhost
       ;

set: SET DEBUG STATE {
		if (conf.debug < 0)
			conf.debug = $3;
   }
   | SET PIDFILE WORD {
		conf.pidfile = $3;
   }
   | SET CHROOT WORD {
		word_is_fs($3, WANT_DIRNAME);
		conf.chroot = $3;
   }
   | SET MIME WORD {
		word_is_fs($3, WANT_FILENAME);
		conf.mimefile = $3;
   }
   | SET USER WORD {
		word_is_login($3);
		conf.user = $3;
   }
   | SET CGI_PATTERN cgi_pattern_content {
   }
   | SET NO_LOG no_log_content {
   }
   | SET INDEX_NAME index_name_content {
   }
   | SET RESTRICT_FROM restrict_from_content {
   }
   | SET ROOTDIR WORD {
		word_is_fs($3, WANT_DIRNAME);
		vh->rootdir = $3;
   }
   | SET DIRLIST_CSS_URL WORD {
		vh->dirlist_css_url = $3;
   }
   | SET CHARSET WORD {
		vh->charset = $3;
   }
   | SET STATUS_URL WORD {
		vh->status_url = $3;
   }
   | SET CGI_ENV cgi_env_content {
   }
   | SET CGI_MAP cgi_map_content {
   }
   | SET MAXAGE NUMBER {
		vh->maxage = $3;
   }
   | SET X_FORWARDED_FOR STATE {
		vh->x_forwarded_for = $3;
   }
   ;

reset: RESET CGI_PATTERN {
		struct mohawk_list *current;
		struct mohawk_list *next;
		current = TAILQ_FIRST(&vh->cgi_patterns);
		while (current != NULL) {
			next = TAILQ_NEXT(current, entry);
			free(current);
			current = next;
		}
		TAILQ_INIT(&vh->cgi_patterns);
   }
   | RESET NO_LOG {
		struct mohawk_list *current;
		struct mohawk_list *next;
		current = TAILQ_FIRST(&vh->no_log_patterns);
		while (current != NULL) {
			next = TAILQ_NEXT(current, entry);
			free(current);
			current = next;
		}
		TAILQ_INIT(&vh->no_log_patterns);
   }
   | RESET INDEX_NAME {
		struct mohawk_list *current;
		struct mohawk_list *next;
		current = TAILQ_FIRST(&vh->index_names);
		while (current != NULL) {
			next = TAILQ_NEXT(current, entry);
			free(current);
			current = next;
		}
		TAILQ_INIT(&vh->index_names);
   }
   | RESET RESTRICT_FROM {
		struct mohawk_list *current;
		struct mohawk_list *next;
		current = TAILQ_FIRST(&vh->restrict_froms);
		while (current != NULL) {
			next = TAILQ_NEXT(current, entry);
			free(current);
			current = next;
		}
		TAILQ_INIT(&vh->restrict_froms);
   }
   | RESET CGI_ENV {
		struct mohawk_hash *current;
		struct mohawk_hash *next;
		current = TAILQ_FIRST(&vh->cgi_env);
		while (current != NULL) {
			next = TAILQ_NEXT(current, entry);
			free(current);
			current = next;
		}
		TAILQ_INIT(&vh->cgi_env);
   }
   | RESET CGI_MAP {
		struct mohawk_hash *current;
		struct mohawk_hash *next;
		current = TAILQ_FIRST(&vh->cgi_map);
		while (current != NULL) {
			next = TAILQ_NEXT(current, entry);
			free(current);
			current = next;
		}
		TAILQ_INIT(&vh->cgi_map);
   }
   | RESET MAXAGE {
		vh->maxage = -1;
   }
   ;

listen: LISTEN STATE WORD PORT NUMBER {
			number_is_port($5);
			char buf[6];
			snprintf(buf, sizeof(buf), "%d", $5);
			if (interface($3, buf) != 1 && host($3, buf) < 0) {
				yyerror("invalid virtual ip or interface or port: '%s' [%d]", $3, $5);
				YYERROR;
			}
        }
      | LISTEN STATE WORD PORT WORD {
			if (interface($3, $5) != 1 && host($3, $5) < 0) {
				yyerror("invalid virtual ip or interface or port: '%s' [%s]", $3, $5);
				YYERROR;
			}
      }
      | LISTEN STATE PORT NUMBER {
			number_is_port($4);
			char buf[6];
			snprintf(buf, sizeof(buf), "%d", $4);
			if (host("0.0.0.0", buf) < 0 || host("::", buf) < 0) {
				yyerror("invalid port: '%d'", $4);
				YYERROR;
			}
      }
      | LISTEN STATE PORT WORD {
			if (host("0.0.0.0", $4) < 0 || host("::", $4) < 0) {
				yyerror("invalid port: '%s'", $4);
				YYERROR;
			}
      }
      | LISTEN STATE WORD {
			if (interface($3, "http") != 1 && host($3, "http") < 0) {
				yyerror("invalid virtual ip or interface or hostname: '%s'", $3);
				YYERROR;
			}
      }
      ;
vhost: VHOST DEFAULT vhost_content {
		add_vhost("default");
     }
     | VHOST WORD vhost_content {
		add_vhost($2);
     }
     ;

cgi_map_content: OBRACE cgi_map_items EBRACE {
              }
              ;

cgi_map_items:
            |
            cgi_map_items cgi_map_item
            ;

cgi_map_item: WORD WORD {
				cgi_map->key = $1;
				cgi_map->value = $2;
				TAILQ_INSERT_TAIL(&vh->cgi_map, cgi_map, entry);
				if ((cgi_map = malloc(sizeof(struct mohawk_hash))) == NULL)
					errx(1, "Unable to allocate memory");
           }
           ;

cgi_pattern_content: OBRACE cgi_pattern_items EBRACE {
                  }
                  ;

cgi_pattern_items:
                |
                cgi_pattern_items cgi_pattern_item
                ;

cgi_pattern_item: WORD {
					cgi_pattern->item = $1;
					TAILQ_INSERT_TAIL(&vh->cgi_patterns, cgi_pattern, entry);
					if ((cgi_pattern = malloc(sizeof(struct mohawk_list))) == NULL)
						errx(1, "Unable to allocate memory");
               }
               ;

no_log_content: OBRACE no_log_items EBRACE {
                  }
                  ;

no_log_items:
                |
                no_log_items no_log_item
                ;

no_log_item: WORD {
					no_log_pattern->item = $1;
					TAILQ_INSERT_TAIL(&vh->no_log_patterns, no_log_pattern, entry);
					if ((no_log_pattern = malloc(sizeof(struct mohawk_list))) == NULL)
						errx(1, "Unable to allocate memory");
               }
               ;

index_name_content: OBRACE index_name_items EBRACE {
                  }
                  ;

index_name_items:
                |
                index_name_items index_name_item
                ;

index_name_item: WORD {
                    index_name->item = $1;
                    TAILQ_INSERT_TAIL(&vh->index_names, index_name, entry);
                    if ((index_name = malloc(sizeof(struct mohawk_list))) == NULL)
                        errx(1, "Unable to allocate memory");
               }
               ;

restrict_from_content: OBRACE restrict_from_items EBRACE {
                  }
                  ;

restrict_from_items:
                |
                restrict_from_items restrict_from_item
                ;

restrict_from_item: WORD {
                    restrict_from->item = $1;
                    TAILQ_INSERT_TAIL(&vh->restrict_froms, restrict_from, entry);
                    if ((restrict_from = malloc(sizeof(struct mohawk_list))) == NULL)
                        errx(1, "Unable to allocate memory");
               }
               | WORDS {
                    restrict_from->item = $1;
                    TAILQ_INSERT_TAIL(&vh->restrict_froms, restrict_from, entry);
                    if ((restrict_from = malloc(sizeof(struct mohawk_list))) == NULL)
                        errx(1, "Unable to allocate memory");
               }
               ;

vhost_content: OBRACE vhost_items EBRACE {
             }
             ;

vhost_items:
           |
           vhost_items vhost_item
           ;

/* We can find reset only inside a vhost block */
vhost_item: set | reset
          ;

cgi_env_content: OBRACE cgi_env_items EBRACE {
           }

cgi_env_items:
         |
         cgi_env_items cgi_env_item
         ;

cgi_env_item: WORD WORD {
			add_cgi_env($1, $2);
        }
        | WORD WORDS {
			add_cgi_env($1, $2);
        }
%%

/* vim: set sw=4 sts=4 ts=4 : */
